from IPython.core.display import HTML
from IPython.core.display import display


def highlight_text(explanation,
                   original_text,
                   show_plot=True,
                   output_html_filename=None,
                   max_opacity=.8):
    """
    Highlights text
    Args:
        explanation: list of tuples of (word, index of word in original data, importance)
        original_text: original text
        show_plot: Shows plot if true (for testing or writing plots to disk instead)
        output_html_filename: Name of the file to save the plot to (optional).
        max_opacity: Maximum opacity (0-1)

    Returns:
        None
    """
    output = _create_html(original_text, explanation, max_opacity)

    if output_html_filename:
        with open(output_html_filename, 'w', encoding='utf-8') as output_html_file:
            print(output, file=output_html_file)

    if show_plot:
        display(HTML, output)


def _create_html(original_text, explanation, max_opacity):
    max_importance = max([abs(item[2]) for item in explanation])
    body = original_text
    words_in_reverse_order = sorted(explanation, key=lambda item: item[1], reverse=True)
    for word, word_start, importance in words_in_reverse_order:
        word_end = word_start + len(word)
        highlighted_word = _highlight_word(word, importance, max_importance, max_opacity)
        body = body[:word_start] + highlighted_word + body[word_end:]
    return '<html><body>' + body + '</body></html>'


def _highlight_word(word, importance, max_importance, max_opacity):
    opacity = max_opacity * abs(importance) / max_importance
    if importance > 0:
        color = f'rgba(255, 0, 0, {opacity:.2f})'
    else:
        color = f'rgba(0, 0, 255, {opacity:2f})'
    highlighted_word = f'<span style="background:{color}">{word}</span>'
    return highlighted_word



from lime.lime_image import LimeImageExplainer
from lime.lime_text import LimeTextExplainer
from dianna.utils import get_function


class LIME:
    """
    LIME implementation as wrapper around https://github.com/marcotcr/lime
    """

    def __init__(self,
                 kernel_width=25,
                 kernel=None,
                 verbose=False,
                 class_names=None,
                 feature_selection='auto',
                 split_expression=r'\W+',
                 bow=False,
                 mask_string=None,
                 random_state=None,
                 char_level=False,
                 preprocess_function=None,
                 ):  # pylint: disable=too-many-arguments
        """LIME initializer.

        Args:
            kernel_width (int, optional):
            kernel (callable, optional):
            verbose (bool, optional):
            class_names (list, optional):
            feature_selection (str, optional):
            split_expression (regexp, optional):
            bow (bool, optional):
            mask_string (str, optional):
            random_state (int or np.RandomState, optional):
            char_level (bool, optional):
            preprocess_function (callable, optional): Function to preprocess input data with
        """
        self.text_explainer = LimeTextExplainer(kernel_width,
                                                kernel,
                                                verbose,
                                                class_names,
                                                feature_selection,
                                                split_expression,
                                                bow,
                                                mask_string,
                                                random_state,
                                                char_level,
                                                )

        self.image_explainer = LimeImageExplainer(kernel_width,
                                                  kernel,
                                                  verbose,
                                                  feature_selection,
                                                  random_state,
                                                  )

        self.preprocess_function = preprocess_function

    def explain_text(self,
                     model_or_function,
                     input_data,
                     labels=(0,),
                     top_labels=None,
                     num_features=10,
                     num_samples=5000,
                     distance_metric='cosine',
                     model_regressor=None,
                     ):  # pylint: disable=too-many-arguments
        """
        Run the LIME explainer.

        Args:
            model_or_function (callable or str): The function that runs the model to be explained _or_
                                                 the path to a ONNX model on disk.
            input_data (np.ndarray): Data to be explained
            labels ([int], optional): Iterable of indices of class to be explained
            top_labels ([type], optional):
            num_features (int, optional):
            num_samples (int, optional):
            distance_metric (str, optional):
            model_regressor ([type], optional):

        Returns:
            list of (word, index of word in raw text, importance for target class) tuples
        """
        runner = get_function(model_or_function, preprocess_function=self.preprocess_function)
        explanation = self.text_explainer.explain_instance(input_data,
                                                           runner,
                                                           labels,
                                                           top_labels,
                                                           num_features,
                                                           num_samples,
                                                           distance_metric,
                                                           model_regressor,
                                                           )

        local_explanations = explanation.local_exp
        string_map = explanation.domain_mapper.indexed_string
        return [self._get_results_for_single_label(local_explanations[label], string_map) for label in labels]

    @staticmethod
    def _get_results_for_single_label(local_explanation, string_map):
        return [(string_map.word(index), int(string_map.string_position(index)), importance)
                for index, importance in local_explanation]

    def explain_image(self,
                      model_or_function,
                      input_data,
                      label=1,
                      hide_color=None,
                      top_labels=None,
                      num_features=10,
                      num_samples=5000,
                      batch_size=10,
                      segmentation_fn=None,
                      distance_metric='cosine',
                      model_regressor=None,
                      random_seed=None,
                      ):  # pylint: disable=too-many-arguments,too-many-locals
        """
        Run the LIME explainer.

        Args:
            model_or_function (callable or str): The function that runs the model to be explained _or_
                                                 the path to a ONNX model on disk.
            input_data (np.ndarray): Data to be explained
            label (int): Index of class to be explained
            hide_color (float, optional):
            top_labels (int, optional):
            num_features (int, optional):
            num_samples (int, optional):
            batch_size (int, optional):
            segmentation_fn (callable, optional):
            distance_metric (str, optional):
            model_regressor ([type], optional):
            random_seed (int, optional):

        Returns:
            list of (word, index of word in raw text, importance for target class) tuples
        """
        runner = get_function(model_or_function, preprocess_function=self.preprocess_function)
        # remove batch axis from input data; this is only here for a consistent API
        # but LIME wants data without batch axis
        if not len(input_data) == 1:
            raise ValueError("Length of batch axis must be one.")
        input_data = input_data[0]
        explanation = self.image_explainer.explain_instance(input_data,
                                                            runner,
                                                            labels=(label,),
                                                            hide_color=hide_color,
                                                            top_labels=top_labels,
                                                            num_features=num_features,
                                                            num_samples=num_samples,
                                                            batch_size=batch_size,
                                                            segmentation_fn=segmentation_fn,
                                                            distance_metric=distance_metric,
                                                            model_regressor=model_regressor,
                                                            random_seed=random_seed
                                                            )

        mask = explanation.get_image_and_mask(label, positive_only=False, hide_rest=True, num_features=num_features)[1]
        return mask



import numpy as np
from skimage.transform import resize
from tqdm import tqdm
from dianna.utils import get_function


def normalize(saliency, n_masks, p_keep):
    return saliency / n_masks / p_keep


class RISE:
    """
    RISE implementation based on https://github.com/eclique/RISE/blob/master/Easy_start.ipynb
    """

    def __init__(self, n_masks=1000, feature_res=8, p_keep=0.5, preprocess_function=None, ):
        """RISE initializer.

        Args:
            n_masks (int): Number of masks to generate.
            feature_res (int): Resolution of features in masks.
            p_keep (float): Fraction of image to keep in each mask
            preprocess_function (callable, optional): Function to preprocess input data with
        """
        self.n_masks = n_masks
        self.feature_res = feature_res
        self.p_keep = p_keep
        self.preprocess_function = preprocess_function
        self.masks = None
        self.predictions = None

    def explain_text(self, model_or_function, input_text, labels=(0,), batch_size=100):
        runner = get_function(model_or_function, preprocess_function=self.preprocess_function)
        input_tokens = np.asarray(model_or_function.tokenizer(input_text))
        text_length = len(input_tokens)
        self.masks = self._generate_masks_for_text(text_length)  # Expose masks for to make user inspection possible
        sentences = self._create_masked_sentences(input_tokens)
        saliencies = self._get_saliencies(runner, sentences, text_length, batch_size)
        return self._reshape_result(input_tokens, labels, saliencies)

    @staticmethod
    def _reshape_result(input_tokens, labels, saliencies):
        word_lengths = [len(t) for t in input_tokens]
        word_indices = [sum(word_lengths[:i]) + i for i in range(len(input_tokens))]
        return [list(zip(input_tokens, word_indices, saliencies[label])) for label in labels]

    def _get_saliencies(self, runner, sentences, text_length, batch_size):
        self.predictions = self._get_predictions(sentences, runner, batch_size)
        unnormalized_saliency = self.predictions.T.dot(self.masks.reshape(self.n_masks, -1)).reshape(-1, text_length)
        return normalize(unnormalized_saliency, self.n_masks, self.p_keep)

    def _get_predictions(self, sentences, runner, batch_size):
        predictions = []
        for i in tqdm(range(0, self.n_masks, batch_size), desc='Explaining'):
            predictions.append(runner(sentences[i:i + batch_size]))
        predictions = np.concatenate(predictions)
        return predictions

    def _create_masked_sentences(self, tokens):
        tokens_masked = []
        for mask in self.masks:
            tokens_masked.append(tokens[mask])
        sentences = [" ".join(t) for t in tokens_masked]
        return sentences

    def _generate_masks_for_text(self, input_size):

        masks = np.random.choice(a=(True, False), size=(self.n_masks, input_size), p=(self.p_keep, 1 - self.p_keep))
        return masks

    def explain_image(self, model_or_function, input_data, batch_size=100):
        """Run the RISE explainer.
           The model will be called with masked images,
           with a shape defined by `batch_size` and the shape of `input_data`

        Args:
            model_or_function (callable or str): The function that runs the model to be explained _or_
                                                 the path to a ONNX model on disk.
            input_data (np.ndarray): Image to be explained
            batch_size (int): Batch size to use for running the model.

        Returns:
            Explanation heatmap for each class (np.ndarray).
        """
        runner = get_function(model_or_function, preprocess_function=self.preprocess_function)

        # data shape without batch axis and (optional) channel axis
        img_shape = input_data.shape[1:3]
        self.masks = self.generate_masks_for_images(img_shape)  # Expose masks for to make user inspection possible

        predictions = []

        # Make sure multiplication is being done for correct axes
        masked = input_data * self.masks

        for i in tqdm(range(0, self.n_masks, batch_size), desc='Explaining'):
            predictions.append(runner(masked[i:i + batch_size]))
        predictions = np.concatenate(predictions)
        self.predictions = predictions
        saliency = predictions.T.dot(self.masks.reshape(self.n_masks, -1)).reshape(-1, *img_shape)
        return normalize(saliency, self.n_masks, self.p_keep)

    def generate_masks_for_images(self, input_size):
        """Generate a set of random masks to mask the input data

        Args:
            input_size (int): Size of a single sample of input data, for images without the channel axis.
        Returns:
            The generated masks (np.ndarray)
        """
        cell_size = np.ceil(np.array(input_size) / self.feature_res)
        up_size = (self.feature_res + 1) * cell_size

        grid = np.random.choice(a=(True, False), size=(self.n_masks, self.feature_res, self.feature_res),
                                p=(self.p_keep, 1 - self.p_keep))
        grid = grid.astype('float32')

        masks = np.empty((self.n_masks, *input_size))

        for i in tqdm(range(self.n_masks), desc='Generating masks'):
            y = np.random.randint(0, cell_size[0])
            x = np.random.randint(0, cell_size[1])
            # Linear upsampling and cropping
            masks[i, :, :] = self._upscale(grid[i], up_size)[y:y + input_size[0], x:x + input_size[1]]
        masks = masks.reshape(-1, *input_size, 1)
        return masks

    def _upscale(self, grid_i, up_size):
        return resize(grid_i, up_size, order=1, mode='reflect', anti_aliasing=False)



import numpy as np
import onnxruntime as ort


class SimpleModelRunner:
    def __init__(self, filename, preprocess_function=None):
        """Generates function to run ONNX model with one set of inputs and outputs

            Args:
                filename (str): Path to ONNX model on disk
                preprocess_function (callable, optional): Function to preprocess input data with

            Returns:
                function

            Examples:
                >>> runner = SimpleModelRunner('path_to_model.onnx')
                >>> predictions = runner(input_data)
        """
        self.filename = filename
        self.preprocess_function = preprocess_function

    def __call__(self, input_data):
        # get ONNX predictions
        sess = ort.InferenceSession(self.filename)
        input_name = sess.get_inputs()[0].name
        output_name = sess.get_outputs()[0].name

        if self.preprocess_function is not None:
            input_data = self.preprocess_function(input_data)

        onnx_input = {input_name: input_data.astype(np.float32)}
        pred_onnx = sess.run([output_name], onnx_input)[0]
        return pred_onnx




from dianna.utils.onnx_runner import SimpleModelRunner


def get_function(model_or_function, preprocess_function=None):
    """Converts input to callable function.
       Input can be either model path or function.
       If input is a function, the function is returned unchanged.

       Any keyword arguments are given to the ModelRunner class if the input is a model path.
    """
    if isinstance(model_or_function, str):
        runner = SimpleModelRunner(model_or_function, preprocess_function=preprocess_function)
    elif callable(model_or_function):
        if preprocess_function is None:
            runner = model_or_function
        else:
            def runner(input_data):
                return model_or_function(preprocess_function(input_data))
    else:
        raise TypeError("model_or_function argument must be string (path to model) or function")
    return runner



import matplotlib.pyplot as plt


def _determine_vmax(max_data_value):
    vmax = 1
    if max_data_value > 255:
        vmax = None
    elif max_data_value > 1:
        vmax = 255
    return vmax


def plot_image(heatmap, original_data=None, heatmap_cmap=None, data_cmap=None, show_plot=True, output_filename=None):  # pylint: disable=too-many-arguments
    """
    Plots a heatmap image
    Args:
        heatmap:
        original_data:
        heatmap_cmap:
        data_cmap:
        show_plot: Shows plot if true (for testing or writing plots to disk instead)
        output_filename: Name of the file to save the plot to (optional).

    Returns:
        None
    """
    # default cmap depends on shape: grayscale or colour

    _, ax = plt.subplots()
    alpha = 1
    if original_data is not None:
        if len(original_data.shape) == 2 and data_cmap is None:
            # 2D array, grayscale
            data_cmap = 'gray'

        ax.imshow(original_data, cmap=data_cmap, vmin=0, vmax=_determine_vmax(original_data.max()))
        alpha = .5

    ax.imshow(heatmap, cmap=heatmap_cmap, alpha=alpha)
    if show_plot:
        plt.show()
    if output_filename:
        plt.savefig(output_filename)
